// $Id: PDF.java 10959 2009-09-29 13:28:31Z chris $
package org.faceless.pdf;

import java.io.*;
import java.util.*;

/**
 * <p>A <code>PDF</code> describes a single document in Adobe's Portable
 * Document Format. It is the highest-level object in the package.</p>
 *
 * <p>
 * The life-cycle of a PDF generally consists of being created, adding
 * new pages, optionally adding information about the document structure
 * (like bookmarks), and finally rendering to an <code>OutputStream</code>.
 * </p><p>
 * This class only deals with the structure of the document. To actually
 * create some content see the {@link PDFPage} class.
 * </p>
 *
 * Here's the ubiquitous example:
 * <pre>
 *   import org.faceless.pdf.*;
 *
 *   // Create a new PDF
 *   PDF p = new PDF();
 *
 *   // Create a new page
 *   PDFPage page = p.newPage(PDF.PAGESIZE_A4);
 *
 *   // Create a new "style" to write in - Black 24pt Times Roman.
 *   PDFStyle mystyle = new PDFStyle();
 *   mystyle.setFont(new StandardFont(StandardFont.TIMES), 24);
 *   mystyle.setFillColor(java.awt.Color.black);
 *
 *   // Put something on the page
 *   page.setStyle(mystyle);
 *   page.drawText("Hello, PDF-viewing World!", 100, 100);
 *
 *   // Automatically go to this page when the document is opened.
 *   p.setOpenAction(PDFAction.goTo(page));
 *
 *   // Add some document info
 *   p.setInfo("Author", "Joe Bloggs");
 *   p.setInfo("Title", "My Document");
 *
 *   // Add a bookmark
 *   java.util.List bookmarks = p.getBookmarks();
 *   bookmarks.add(new PDFBookmark("Hello World page", PDFAction.goTo(page)));
 *
 *   // Allow the user to print the document, but nothing else
 *   p.setAccessLevel(PDF.ACCESS_PRINT);
 *
 *   // Write the document to a file
 *   p.render(new FileOutputStream("test.pdf"));
 * </pre>
 *
 * @see PDFPage
 * @see PDFReader
 * @see PDFBookmark
 * @see PDFStyle
 * @version $Revision: 1.9 $
 *
 */
public class PDF
{
    private org.faceless.pdf2.PDF pdf;

    /**
     * The package version number
     */
    public static final String VERSION="BFOPDF $Revision: 1.9 $";

    /**
     * Argument to <code>setLayout</code> to display each page separately (the default)
     */
    public static final String SINGLEPAGE = "SinglePage";

    /**
     * Argument to <code>setLayout</code> to display the pages in a single column.
     */
    public static final String ONECOLUMN = "OneColumn";

    /**
     * Argument to <code>setLayout</code> to display the pages in two columns, with
     * odd numbered pages to the left.
     */
    public static final String TWOCOLUMNLEFT = "TwoColumnLeft";

    /**
     * Argument to <code>setLayout</code> to display the pages in two columns, with
     * odd numbered pages to the right.
     */
    public static final String TWOCOLUMNRIGHT = "TwoColumnRight";

    /**
     * Argument to <code>newPage</code> to create a new A3 portrait page
     */
    public static final String PAGESIZE_A3 = "842x1190";

    /**
     * Argument to <code>newPage</code> to create a new A3 landscape page
     */
    public static final String PAGESIZE_A3_LANDSCAPE = "1190x842";

    /**
     * Argument to <code>newPage</code> to create a new A4 portrait page
     */
    public static final String PAGESIZE_A4 = "595x842";

    /**
     * Argument to <code>newPage</code> to create a new A4 landscape page
     */
    public static final String PAGESIZE_A4_LANDSCAPE = "842x595";

    /**
     * Argument to <code>newPage</code> to create a new A5 portrait page
     */
    public static final String PAGESIZE_A5 = "421x595";

    /**
     * Argument to <code>newPage</code> to create a new A5 landscape page
     */
    public static final String PAGESIZE_A5_LANDSCAPE = "595x421";

    /**
     * Argument to <code>newPage</code> to create a new A6 portrait page
     */
    public static final String PAGESIZE_A6 = "297x421";

    /**
     * Argument to <code>newPage</code> to create a new A6 landscape page
     */
    public static final String PAGESIZE_A6_LANDSCAPE = "421x297";

    /**
     * Argument to <code>newPage</code> to create a new "letter" portrait page
     */
    public static final String PAGESIZE_LETTER = "612x792";

    /**
     * Argument to <code>newPage</code> to create a new "letter" landscape page
     */
    public static final String PAGESIZE_LETTER_LANDSCAPE = "792x612";

    /**
     * Argument to <code>newPage</code> to create a new "legal" portrait page
     */
    public static final String PAGESIZE_LEGAL = "612x1008";

    /**
     * Argument to <code>newPage</code> to create a new "legal" landscape page
     */
    public static final String PAGESIZE_LEGAL_LANDSCAPE = "1008x612";

    /**
     * Argument to <code>setAccessLevel</code> to allow the user full access
     * to the document.
     * @since 1.1.12
     */
    public static final int ACCESS_ALL = 0xFFFFFFFF;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to print
     * the document. If 128-bit encryption is used, this flag on it's own
     * will only allow low-quality printing - you should set the
     * {@link #ACCESS_HQPRINT} flag too for high quality printing. If
     * 40-bit encryption is used, printing is always high-quality.
     */
    public static final int ACCESS_PRINT = 4;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to edit
     * the document with software like Adobe Acrobat&trade;
     */
    public static final int ACCESS_EDIT = 8;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to copy
     * and paste text and images from the document
     */
    public static final int ACCESS_COPY = 16;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to add
     * annotations to the document and create or edit interactive form
     * fields (including signature fields).
     */
    public static final int ACCESS_ANNOTATE = 32;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to fill
     * in existing interactive form fields (including signature fields),
     * even if ACCESS_ANNOTATE is not set. (128 bit encryption only)
     * @since 1.1.12
     */
    public static final int ACCESS_FORMS = 256;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to
     * extract text and graphics (in support of accessibility for disabled
     * users or for other purposes). (128 bit encryption only)
     * @since 1.1.12
     */
    public static final int ACCESS_EXTRACT = 512;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to
     * assemble the document (insert, rotate or delete pages and create
     * bookmarks or thumbnail images), even if ACCESS_EDIT is not set.
     * (128 bit encryption only)
     * @since 1.1.12
     */
    public static final int ACCESS_ASSEMBLE = 1024;

    /**
     * Argument to <code>setAccessLevel</code> to allow the user to
     * print the document to a representation from which a faithful digital
     * copy of the PDF content could be generated. When this option is not set
     * and ACCESS_PRINT is set, printing is limited to a low-level
     * representation of the appearance, possibly of degraded quality.
     * (128 bit encryption only)
     * @since 1.1.12
     */
    public static final int ACCESS_HQPRINT = 2048;

    /**
     * Argument to <code>setEncryptionAlgorithm</code> to
     * turn off encryption for the document. This is
     * the default.
     * @since 1.1.12
     */
    public static final int ENCRYPT_NONE = 0;

    /**
     * Argument to <code>setEncryptionAlgorithm</code> to
     * set the encryption to the standard 40 bit encryption
     * algorithm used by Acrobat 3.x and later.
     * @since 1.1.12
     */
    public static final int ENCRYPT_40BIT = 40;

    /**
     * Argument to <code>setEncryptionAlgorithm</code> to
     * set the encryption to the 128 bit encryption
     * algorithm added in Acrobat 5.x. Documents encrypted
     * with this algorithm cannot be opened by earlier
     * versions of Acrobat.
     * @since 1.1.12
     */
    public static final int ENCRYPT_128BIT = 128;


    public PDF(org.faceless.pdf2.PDF pdf)
    {
        this.pdf=pdf;
    }

    /**
     * Create a new PDF document.
     */
    public PDF()
    {
        pdf = new org.faceless.pdf2.PDF();
    }

    /**
     * Create a PDF from the specified PDFReader. The <tt>PDFReader</tt> class
     * is available as part of the "Extended Edition" of the PDF library, and is
     * included with this package. If the document contains multiple revisions,
     * the latest revision is loaded.
     * @since 1.1.12
     */
    public PDF(PDFReader reader)
    {
        pdf = new org.faceless.pdf2.PDF(reader.reader);
    }

    /**
     * Create a PDF from the specified PDFReader, using the specified revision of
     * the document. The <tt>PDFReader</tt> class is available as part of the
     * "Extended Edition" of the PDF library, and is included with this package.
     * The revision number must be between 0 and
     * {@link PDFReader#getNumberOfRevisions}, otherwise an
     * <code>IllegalArgumentException</code> is thrown.
     * @param reader the <code>PDFReader</code> to load the document from
     * @param revision the revision of the document to load - betwween
     * <code>reader.getNumberOfRevisions()-1</code> to load the latest revision, or
     * <code>0</code> to load the first.
     * @since 1.2.1
     */
    public PDF(PDFReader reader, int revision)
    {
        pdf = new org.faceless.pdf2.PDF(reader.reader, revision);
    }

    /**
     * Create a new page with the specified pagesize. This version is
     * usually called to set the page to a standard size, like
     * {@link #PAGESIZE_A4}
     * @param pagesize The size of the page as a String of the format "WxH",
     * showing width and height in points.
     */
    public PDFPage newPage(String pagesize)
    {
        return (PDFPage)PeeredObject.getPeer(pdf.newPage(pagesize));
    }

    /**
     * Create a new page in the specified size.
     * @param width The width of the page, in points
     * @param height The height of the page, in points
     */
    public PDFPage newPage(float width, float height)
    {
        return (PDFPage)PeeredObject.getPeer(pdf.newPage((int)width,(int)height));
    }

    /**
     * Return the page returned from the last call to <code>newPage</code>
     * @return the last page returned from the {@link #newPage} method, or
     * <tt>null</tt> if that method hasn't been called yet.
     */
    public PDFPage getLastPage()
    {
        return getPage(getNumberOfPages()-1);
    }

    /**
     * Return the specified page, where <code>0 &lt;= page &lt; numpages</code>.
     * If the page is outside this range, throw an <code>ArrayIndexOutOfBoundsException</code>
     * @return the specified page
     * @throws ArrayIndexOutOfBoundsException if the page is out of range
     * @since 1.1
     */
    public PDFPage getPage(int pagenum)
    throws ArrayIndexOutOfBoundsException
    {
        return (PDFPage)PeeredObject.getPeer(pdf.getPage(pagenum));
    }

    /**
     * Return the number of pages currently defined in the PDF document
     * @since 1.1
     */
    public int getNumberOfPages()
    {
        return pdf.getNumberOfPages();
    }

    /**
     * Return the number of revisions made to the document. This will only be
     * useful for documents read in using a <code>PDFReader</code> - all
     * other PDFs (and many of these too) will return zero. See the
     * {@link PDFReader} class for more information on revisions.
     * @since 1.2.1
     */
    public int getNumberOfRevisions()
    {
        return pdf.getNumberOfRevisions();
    }

    /**
     * Return the list of pages in the PDF document. This list can be manipulated
     * to reorder, delete or append pages as necessary using the standard
     * <tt>java.util.List</tt> methods.
     * @since 1.1.12
     */
    public List getPages()
    {
        /*
         * We need to override these methods because in version 1, moving pages
         * between documents also moved any form fields associated with those
         * pages. This is a bit crude - what happens if a field has annotations
         * on two pages, for instance - but that's what it did, so that's what
         * we do here too.
         */

        return new ListPeer(pdf.getPages())
        {
            public void add(int i, Object o)
            {
                super.add(i,o);
                List l = ((PDFPage)o).page.getAnnotations();
                for (i=0;i<l.size();i++) {
                    if (l.get(i) instanceof org.faceless.pdf2.WidgetAnnotation) {
                        org.faceless.pdf2.WidgetAnnotation annot = (org.faceless.pdf2.WidgetAnnotation)l.get(i);
                        pdf.getForm().getElements().put(annot.getField().getForm().getName(annot.getField()), annot.getField());
                    }
                }
            }

            public Object set(int i, Object o)
            {
                Object old = super.set(i,o);
                List l = ((PDFPage)o).page.getAnnotations();
                for (i=0;i<l.size();i++) {
                    if (l.get(i) instanceof org.faceless.pdf2.WidgetAnnotation) {
                        org.faceless.pdf2.WidgetAnnotation annot = (org.faceless.pdf2.WidgetAnnotation)l.get(i);
                        pdf.getForm().getElements().put(annot.getField().getForm().getName(annot.getField()), annot.getField());
                    }
                }
                return old;
            }
        };
    }

    /**
     * <p>
     * Set the current Locale of the document. The locale language alters how the document
     * processes text internally - for example, setting the locale to
     * <code>Locale.ARABIC</code> sets the default justification to right-aligned.
     * </p><p>
     * The locale may be set many times throughout the document as required. However,
     * changing the locale in the middle of a paragraph should be avoided.
     * </p><p>
     * The default PDF locale is the value of <code>Locale.getDefault()</code>
     * </p>
     * @since 1.1
     */
    public void setLocale(Locale locale)
    {
        pdf.setLocale(locale);
    }

    /**
     * Return the PDF's locale, as set by {@link #setLocale}
     * @since 1.1
     */
    public Locale getLocale()
    {
        return pdf.getLocale();
    }

    /**
     * Controls how the document is initially laid out in the PDF viewer.
     *
     * @param layout How to display the document if there is more than one page.
     * Can be {@link #SINGLEPAGE} (the default), {@link #ONECOLUMN},
     * {@link #TWOCOLUMNLEFT} or {@link #TWOCOLUMNRIGHT}
     * @param bookmarks whether to show bookmarks when the document is first
     * opened. The default is <i>false</i>.
     */
    public void setLayout(String layout, boolean bookmarks)
    {
        pdf.setLayout(layout, bookmarks ? "UseOutlines" : "UseNone");
    }

    /**
     * <p>
     * Add some meta-information on the document, like the subject or title.
     * Although any key or value is valid (except <code>Producer</code>,
     * <code>CreationDate</code> and <code>ModDate</code>, which are set
     * by the library internally), the following fields are recognised
     * by Acrobat:
     * </p><table>
     * <tr><th align=left valign=top>Title</th><td>The document's title</td></tr>
     * <tr><th align=left valign=top>Author</th><td>The name of the person who created the document.</td></tr>
     * <tr><th align=left valign=top>Subject</th><td>The subject of the document.</td></tr>
     * <tr><th align=left valign=top>Keywords</th><td>Keywords associated with the document.</td></tr>
     * <tr><th align=left valign=top>Creator</th><td>If the document was converted to PDF from another format, the
     * name of the application that created the original document from which it was converted.</td></tr>
     * </table>
     * <p>
     * Although both key and value can use Unicode characters outside the
     * ASCII range, whether the viewer can display them correctly is another
     * matter.
     * </p>
     * <p>
     * Note that this meta-information can be set independently of any XML
     * meta information as set by {@link #setMetaData}.
     * </p>
     * @param key the field to set
     * @param val the value to set.
     */
    public void setInfo(String key, String val)
    {
        pdf.setInfo(key,val);
    }

    /**
     * <p>
     * Return document meta information, as set by {@link #setInfo}.
     * </p>
     * <p>
     * Note that this meta-information can be used independently of any XML
     * meta information as returned by {@link #getMetaData}.
     * </p>
     * @param key the field to get
     * @return the value of the specified field, or <tt>null</tt> if the field is not set
     */
    public String getInfo(String key)
    {
        return pdf.getInfo(key);
    }

    /**
     * <p>
     * Return document meta information, as set by {@link #setInfo}. This
     * method returns a map containing all the meta information. If no
     * meta information is available, returns an empty Map.
     * </p><p>
     * Note that this meta-information can be used independently of any XML
     * meta information as returned by {@link #getMetaData}.
     * </p>
     * @return a Map containing any meta information specified in the document.
     * The keys in the Map are Strings, and the values are always Strings or
     * Dates.
     * @since 1.1.12
     */
    public Map getInfo()
    {
        // We have to filter out anything that's not a String or a Date,
        // as that's all v1 functions were expecting back.
        //
        Map m = pdf.getInfo();
        Map out = new HashMap();
        for (Iterator i = m.entrySet().iterator();i.hasNext();) {
            Map.Entry entry = (Map.Entry)i.next();
            if (entry.getValue() instanceof String || entry.getValue() instanceof Date) {
                out.put(entry.getKey(), entry.getValue());
            }
        }
        return Collections.unmodifiableMap(out);
    }

    /**
     * <p>
     * Set the page to open Full Screen or not. This is off by default, and will
     * probably be ignored by any PDF viewer other than Adobe's own Acrobat.
     * </p>
     * @since 1.1.2
     */
    public void setOpenFullScreen(boolean fullscreen)
    {
        pdf.setViewerPreference("FullScreen", fullscreen);
    }

    /**
     * <p>
     * Set the XML Metadata associated with this document. This is a PDF
     * 1.4 feature, so it will only be used by Acrobat 5.0 or later - although
     * it <i>should</i> be ignored by earlier viewers.
     * </p><p>
     * The value is specified as a String, but as it's XML it will probably be
     * created as a <tt>java.io.Writer</tt>. Here's an example showing how
     * to use this with SAX and the <tt>org.apache.xml.serialize</tt>
     * classes. The lines marked in bold could apply to any method of
     * serializing the XML.
     * </p>
     * <pre>
     *   void addMetaData(PDF pdf, InputSource source)
     *   {
     *       <i>// Create a java.io.Writer</i>
     *       <b>StringWriter meta = new StringWriter();</b>
     *
     *       <i>// Create a SAX XML serializer using the apache classes</i>
     *       <i>// We must not include the XML Declaration in the output.</i>
     *       OutputFormat outformat = new OutputFormat();
     *       outformat.setOmitXMLDeclaration(true);
     *       Serializer serial = new XMLSerializer(out, outformat);
     *
     *       <i>// Create the XML parser and parse the input source</i>
     *       XMLReader parser = XMLReaderFactory.createXMLReader();
     *       parser.setContentHandler(serial.asContentHandler());
     *       parser.parse(inputsource);
     *
     *       <b>pdf.setMetaData(meta.toString());</b>
     *   }
     * </pre>
     * <p>
     * For more information on XML metadata in PDF documents, see
     * <a href="http://www.adobe.com/products/xmp">http://www.adobe.com/products/xmp</a>
     * </p>
     * @param xmldata the XML data to embed into the document. No validation is performed
     * on this input.
     * @since 1.1.12
     */
    public void setMetaData(String xmldata)
    {
        pdf.setMetaData(xmldata);
    }

    /**
     * <p>
     * Return any XML Metadata associated with the document. XML Metadata
     * is a PDF 1.4 (Acrobat 5.x) feature, so this method will return
     * <tt>null</tt> for all documents matching earlier specifications or
     * if no metadata is specified.
     * </p><p>
     * Here's an example of how to extract the MetaData into a DOM tree.
     * Lines in bold could apply to any parsing method.
     * </p>
     * <pre>
     *   <b>InputSource source = new InputSource(pdf.getMetaData());</b>
     *   DOMParser parser = new DOMParser();
     *   parser.parse(source);
     *   Document doc = parser.getDocument();
     * </pre>
     * <p>
     * For more information on XML metadata in PDF documents, see
     * <a href="http://www.adobe.com/products/xmp">http://www.adobe.com/products/xmp</a>
     * </p>
     * @since 1.1.12
     * @returns a {@link java.io.Reader} containing the source of the XML
     */
    public Reader getMetaData()
    throws IOException
    {
        return pdf.getMetaData();
    }

    /**
     * <p>
     * Set some preferences as to how the document is displayed on screen. Along with the
     * <code>setLayout</code> function (which set some of the
     * more conventional viewer preferences), this function allow you to turn on or off some
     * of the more obscure features of the users PDF viewer application. By default, all
     * these options are set to <i>false</i>. Valid options are:
     * </p>
     * <table>
     * <tr><th align=left valign=top>HideToolbar</th><td>A flag specifying whether to hide the viewer application's tool bars when the document is active.</td></tr>
     * <tr><th align=left valign=top>HideMenubar</th><td>A flag specifying whether to hide the viewer application's menu bar when the document is active.</td></tr>
     * <tr><th align=left valign=top>HideWindowUI</th><td>A flag specifying whether to hide user interface elements in the document' window (such as scroll bars and navigation controls), leaving only the document's contents displayed.</td></tr>
     * <tr><th align=left valign=top>FitWindow</th><td>A flag specifying whether to resize the document's window to fit the size of the first displayed page. Note this resizes the window to fit the document, not the other way round.</td></tr>
     * <tr><th align=left valign=top>CenterWindow</th><td>A flag specifying whether to position the document's window in the center of the screen. Note this moves the whole viewer to the center of the screen, not the document to the center of the viewer.</td></tr>
     * </table>
     * <p>Viewers other than Adobes own Acrobat reader will probably ignore these options.</p>
     * @param key the option to set, from the list above
     * @param val whether to turn that option on or off
     */
    public void setViewerPreference(String key, boolean val)
    {
        pdf.setViewerPreference(key, val);
    }

    /**
     * Return the value of the specified viewer preference, as set by {@link #setViewerPreference}.
     * @param key the option to query
     * @return <tt>true</tt> if the option is set to "on", or <tt>false</tt> if the option is set to "off" or not set.
     * @since 1.1.12
     */
    public boolean getViewerPreference(String key)
    {
        return pdf.getViewerPreference(key);
    }

    /**
     * <p>
     * Set the action to perform when the document is first opened. Use
     * this method to open the PDF at a specific page, or set the initial
     * zoom level.
     * </p><p>
     * To run an action when a specific <i>page</i> is opened, see the
     * method with the same name in the {@link PDFPage} object.
     * </p>
     * @param action The action to perform on opening. Since 1.1, set to
     * <code>null</code> to remove the action.
     */
    public void setOpenAction(PDFAction action)
    {
        pdf.setAction(org.faceless.pdf2.Event.OPEN, action==null ? null : action.action);
    }

    /**
     * Return the action that is performed when the document is first
     * opened.
     * @return the action that is performed when the document is first
     * opened, or <tt>null</tt> if no action is performed
     * @since 1.1.12
     */
    public PDFAction getOpenAction()
    {
        return (PDFAction)PeeredObject.getPeer(pdf.getAction(org.faceless.pdf2.Event.OPEN));
    }

    /**
     * <p>
     * Create a "named" action, which can be referenced from outside the
     * current PDF, useful to (for example) open the generated PDF at a
     * specific position from an HTML link. Here's how to do this:
     * </p>
     * In the PDF, add the following code:
     * <pre>
     *   pdf.setNamedAction("Myaction", PDFAction.goTo(somepage))
     * </pre>
     * Then in your HTML document, add the following code:
     * <pre>
     *   &lt;a href="http://www.mycompany.com/mypdf.pdf#Myaction"&gt;
     * </pre>
     * <p>
     * The action name is case sensitive. Adding a <tt>null</tt> action
     * causes the specified named action to be deleted. The action <i>must</i>
     * be a standard "goTo" action, jumping to a page in the document.
     * </p>
     * @param name The name to give this action.
     * @param action The action to perform on opening, or <tt>null</tt> to
     * delete the action
     * @since 1.1.5
     */
    public void setNamedAction(String name, PDFAction action)
    {
        if (action==null) {
            pdf.getNamedActions().remove(name);
        } else {
            pdf.getNamedActions().put(name, action.action);
        }
    }

    /**
     * Return a Map of all the "named actions" in the document,
     * as set by {@link #setNamedAction}.
     * @return A Map of all the named actions in the document. The key is
     * the {@link java.lang.String} which is the name of the action, the
     * value is the {@link PDFAction} to be performed. If no named actions
     * are specified, an empty map is returned.
     * @since 1.1.12
     */
    public Map getNamedActions()
    {
        // Return a copy, because in v1 changing the returned map has no effect.
        TreeMap map = new TreeMap();
        for (Iterator i = pdf.getNamedActions().entrySet().iterator();i.hasNext();) {
            Map.Entry entry = (Map.Entry)i.next();
            map.put(entry.getKey(), PeeredObject.getPeer(entry.getValue()));
        }
        return map;
    }

    /**
     * Return the list of Bookmarks at the top level of the document. The List
     * contains zero or more elements of type {@link PDFBookmark}. The List can
     * be altered using any of the standard {@link java.util.List} methods to
     * order the documents bookmarks in any way you see fit. By default, all
     * documents start with an empty List.
     * @see PDFBookmark
     * @return The List of bookmarks at the top level of the document
     */
    public List getBookmarks()
    {
        return new ListPeer(pdf.getBookmarks());
    }

    // ---------------------------------------------------------------------------------

    /**
     * <p>
     * Set the "security" password for the PDF document - the password
     * required to change the security settings of the document (the access
     * level and the open password). If you don't anticipate changing
     * the security settings at a later date, you can leave this blank.
     * </p><p>
     * If no encryption type has been set by calling {@link #setEncryptionAlgorithm},
     * the algorithm is set to the standard 40 bit encryption.
     * </p>
     * @since 1.1
     */
    public void setSecurityPassword(String s)
    {
        org.faceless.pdf2.StandardEncryptionHandler handler = (org.faceless.pdf2.StandardEncryptionHandler)pdf.getEncryptionHandler();
        if (s==null) s="";
        if (handler==null) {
            handler = new org.faceless.pdf2.StandardEncryptionHandler();
            pdf.setEncryptionHandler(handler);
        }
        handler.setOwnerPassword(s);
    }

    /**
     * <p>
     * Set the password required to open the PDF document.
     * (also called the "User" password). It can be left blank,
     * in which case the document can be opened by anyone.
     * </p><p>
     * If no encryption type has been set by calling {@link #setEncryptionAlgorithm},
     * the algorithm is set to the standard 40 bit encryption.
     * </p>
     * @since 1.1
     */
    public void setPassword(String s)
    {
        org.faceless.pdf2.StandardEncryptionHandler handler = (org.faceless.pdf2.StandardEncryptionHandler)pdf.getEncryptionHandler();
        if (s==null) s="";
        if (handler==null) {
            handler = new org.faceless.pdf2.StandardEncryptionHandler();
            pdf.setEncryptionHandler(handler);
        }
        handler.setUserPassword(s);
    }

    /**
     * <p>
     * Set the encryption algorithm used to encrypt the document.
     * Currently limited to either {@link #ENCRYPT_NONE} to disable
     * encryption, {@link #ENCRYPT_40BIT} to use the standard 40-bit
     * algorithm accepted by most modern versions of Acrobat, or
     * {@link #ENCRYPT_128BIT} to use the expanded 128-bit algorithm
     * introduced in PDF1.4 (Acrobat 5).
     * </p>
     * @since 1.1.12
     */
    public void setEncryptionAlgorithm(int type)
    {
        if (type==ENCRYPT_NONE) {
            pdf.setEncryptionHandler(null);
        } else {
            org.faceless.pdf2.StandardEncryptionHandler handler = (org.faceless.pdf2.StandardEncryptionHandler)pdf.getEncryptionHandler();
            if (handler==null) {
                handler = new org.faceless.pdf2.StandardEncryptionHandler();
                pdf.setEncryptionHandler(handler);
            }
            if (type==ENCRYPT_128BIT) {
                handler.setAcrobat5Level(handler.PRINT_HIGHRES, handler.EXTRACT_ALL, handler.CHANGE_ALL);
            } else {
                handler.setAcrobat3Level(true, true, true, true);
            }
        }
    }

    /**
     * <p>
     * Set the level of access available to the user. The user can be granted any
     * of the {@link #ACCESS_PRINT}, {@link #ACCESS_EDIT}, {@link #ACCESS_COPY}
     * or {@link #ACCESS_ANNOTATE} rights. Additionally, if 128 bit encryption
     * is used the {@link #ACCESS_FORMS}, {@link #ACCESS_ASSEMBLE}, {@link #ACCESS_EXTRACT}
     * and {@link #ACCESS_HQPRINT} options can be used.
     * </p><p>
     * Please remember it is <i>up to the PDF viewer software</i> to enforce this, and
     * that there are plenty of applications available on the 'net to allow full access
     * to a PDF document. Consequently this setting should be considered a <b>recommendation
     * only</b>.
     * </p><p>
     * Setting this value involves encrypting the document, even if no password has been
     * set. If no encryption type has been set by calling  {@link #setEncryptionAlgorithm},
     * the algorithm is set to the standard 40 bit encryption.
     * </p><p>
     * The default is that the user has full access to the document.
     * </p>
     * @param level a logical-OR of any of the above flags
     * @since 1.1
     */
    public void setAccessLevel(int level)
    {
        org.faceless.pdf2.StandardEncryptionHandler handler = (org.faceless.pdf2.StandardEncryptionHandler)pdf.getEncryptionHandler();
        if (handler==null) {
            handler = new org.faceless.pdf2.StandardEncryptionHandler();
            pdf.setEncryptionHandler(handler);
        }

        if (handler.getVersion()==2) {
            boolean print = (level&ACCESS_PRINT)!=0;
            boolean annotate = (level&ACCESS_ANNOTATE)!=0;
            boolean edit = (level&ACCESS_EDIT)!=0;
            boolean copy = (level&ACCESS_COPY)!=0;
            boolean forms = (level&ACCESS_FORMS)!=0;
            boolean extract = (level&ACCESS_EXTRACT)!=0;
            boolean assemble = (level&ACCESS_ASSEMBLE)!=0;
            boolean hqprint = (level&ACCESS_HQPRINT)!=0;

            int newchange=handler.CHANGE_NONE, newprint=handler.PRINT_NONE, newextract=handler.EXTRACT_NONE;

            if (print && hqprint) {
                newprint=handler.PRINT_HIGHRES;
            } else if (print) {
                newprint=handler.PRINT_LOWRES;
            } else {
                newprint=handler.PRINT_NONE;
            }

            if (extract && copy) {
                newprint=handler.EXTRACT_ALL;
            } else if (extract) {
                newprint=handler.EXTRACT_ACCESSIBILITY;
            } else {
                newprint=handler.EXTRACT_NONE;
            }

            if (assemble && annotate && edit) {
                newchange = handler.CHANGE_ALL;
            } else if (assemble && annotate) {
                newchange = handler.CHANGE_ANNOTATIONS;
            } else if (forms) {
                newchange = handler.CHANGE_FORMS;
            } else if (assemble) {
                newchange = handler.CHANGE_LAYOUT;
            } else {
                newchange = handler.CHANGE_NONE;
            }

            handler.setAcrobat5Level(newprint, newextract, newchange);
        }
        else
        {
            boolean print = (level&ACCESS_PRINT)!=0;
            boolean annotate = (level&ACCESS_ANNOTATE)!=0;
            boolean edit = (level&ACCESS_EDIT)!=0;
            boolean copy = (level&ACCESS_COPY)!=0;
            handler.setAcrobat3Level(print, annotate, copy, edit);
        }
    }


    //--------------------------------------------------------------------


    /**
     * Return the Interactive {@link Form} or "AcroForm" object which is part
     * of each PDF document.
     * <b>Note that using interactive forms requires the "Extended Edition"
     * of the library</b> - although the classes are supplied with the package an "Extended
     * Edition" license must be purchased to activate this functionality.
     * @since 1.1.13
     */
    public Form getForm()
    {
        return (Form)PeeredObject.getPeer(pdf.getForm());
    }

    /**
     * Set the document-wide JavaScript. This JavaScript is executed when
     * the document is first loaded - this is normally used to define
     * functions and the like, in the same way as JavaScript defined in the
     * &lt;HEAD&gt; of an HTML document.
     * @param javascript the JavaScript to use for the entire document
     * @since 1.1.23
     * @see #getJavaScript
     * @see PDFAction#formJavaScript
     */
    public void setJavaScript(String javascript)
    {
        pdf.setJavaScript(javascript);
    }

    /**
     * Return the document-wide JavaScript, as set by {@link #setJavaScript},
     * or <code>null</code> if no JavaScript is defined for this document.
     * @since 1.1.23
     * @see #setJavaScript
     * @see PDFAction#formJavaScript
     */
    public String getJavaScript()
    {
        return pdf.getJavaScript();
    }

    /**
     * <p>
     * This method renders the completed PDF to an OutputStream. The
     * stream is left open on completion. A document may be rendered
     * more than once. Rendering the document merges all the revisions
     * of a document, so after rendering the {@link #getNumberOfRevisions}
     * method will always return zero.
     * </p>
     * @param out the output stream to write the PDF to.
     * @throws IOException if the rendering process could not be completed
     */
    public synchronized void render(OutputStream out)
    throws IOException
    {
        pdf.render(out);
    }

    /**
     * Set the license key for the library. When the library is purchased,
     * the Big Faceless Organization supplies a key which removes the "DEMO"
     * stamp on each of the documents
     * @param key the license key
     */
    public static void setLicenseKey(String key)
    {
        org.faceless.pdf2.PDF.setLicenseKey(key);
    }

    /**
     * Import the contents of the specified {@link FDF} into the PDF document.
     * Although the FDF file may contain a wide variety of information, the
     * only part of it actually used by this method are the form values, which
     * are used to set the corresponding form fields in the PDF. If a field
     * doesn't exist, a warning is printed and the field is ignored
     * @since 1.2.1
     */
    public void importFDF(FDF fdf)
    {
        pdf.importFDF((org.faceless.pdf2.FDF)fdf.fdf);
    }
}
